import logging
from setuptools import setup, find_packages

tutor_deps = [
    "pillow",
    "tqdm",
    "ipython"
]
test_deps = [
    'pytest>=4',
    'pytest-cov>=2.6.0',
    'pytest-flake8',
]
docs_deps = [
    'sphinx',
    'sphinx_rtd_theme',
    'sphinx_toggleprompt',
    'sphinx-gallery>=0.6',
    'nbsphinx',
    'm2r2'
]

dev_deps = ["requests"] + docs_deps + test_deps

try:
    import torch

    ml_pytorch_deps = []
except ModuleNotFoundError:
    import sys

    if 5 <= sys.version_info[1]:
        ml_pytorch_deps = ["torch"]
    else:
        ml_pytorch_deps = []
        logging.warning("Current python version %s is not supported by pytorch", str(sys.version_info[:2]))

vec_deps = ['gensim'] + ml_pytorch_deps

setup(
    name='EduNLP',
    version='0.0.6',
    extras_require={
        'test': test_deps,
        'doc': docs_deps,
        'tutor': tutor_deps,
        'dev': dev_deps,
        'vec': vec_deps,
        'full': vec_deps + tutor_deps
    },
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'networkx',
        'numpy>=1.17.0',
        'jieba',
        'js2py',
        'EduData>=0.0.16',
        'PyBaize>=0.0.3'
    ],  # And any other dependencies foo needs
    entry_points={
        "console_scripts": [
            "edunlp = EduNLP.main:cli",
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# coding: utf-8
# 2021/5/29 @ tongshiwei

from .gensim_vec import W2V, D2V, BowLoader, TfidfLoader
from .const import *
from .rnn import RNNModel
from .t2v import T2V, get_pretrained_t2v
from .embedding import Embedding



# coding: utf-8
# 2021/7/13 @ tongshiwei

import os
from longling import path_append
from EduData import get_data
from .rnn import RNNModel
from .gensim_vec import W2V, D2V
from .meta import Vector
from EduNLP.constant import MODEL_DIR

MODELS = {
    "w2v": W2V,
    "d2v": D2V,
    "rnn": RNNModel,
    "lstm": RNNModel,
    "gru": RNNModel,
    "elmo": RNNModel
}


class T2V(object):
    def __init__(self, model: str, *args, **kwargs):
        model = model.lower()
        self.model_type = model
        if model in {"rnn", "lstm", "gru", "elmo"}:
            self.i2v: Vector = MODELS[model](model, *args, **kwargs)
        else:
            self.i2v: Vector = MODELS[model](*args, **kwargs)

    def __call__(self, items, *args, **kwargs):
        return self.i2v.infer_vector(items, *args, **kwargs)

    def infer_vector(self, items, *args, **kwargs):
        return self.i2v.infer_vector(items, *args, **kwargs)

    def infer_tokens(self, items, *args, **kwargs):
        return self.i2v.infer_tokens(items, *args, **kwargs)

    @property
    def vector_size(self) -> int:
        return self.i2v.vector_size


PRETRAINED_MODELS = {
    "d2v_all_256": ["http://base.ustc.edu.cn/data/model_zoo/EduNLP/d2v/general_all_256.zip", "d2v"],
    "d2v_sci_256": ["http://base.ustc.edu.cn/data/model_zoo/EduNLP/d2v/general_science_256.zip", "d2v"],
    "d2v_eng_256": ["http://base.ustc.edu.cn/data/model_zoo/EduNLP/d2v/general_english_256.zip", "d2v"],
    "d2v_lit_256": ["http://base.ustc.edu.cn/data/model_zoo/EduNLP/d2v/general_literal_256.zip", "d2v"],
    "w2v_eng_300": ["http://base.ustc.edu.cn/data/model_zoo/EduNLP/w2v/general_english_300.zip", "w2v"],
    "w2v_lit_300": ["http://base.ustc.edu.cn/data/model_zoo/EduNLP/w2v/general_literal_300.zip", "w2v"],
}


def get_pretrained_t2v(name, model_dir=MODEL_DIR):
    if name not in PRETRAINED_MODELS:
        raise KeyError(
            "Unknown pretrained model %s, use one of the provided pretrained models: %s" % (
                name, ", ".join(PRETRAINED_MODELS.keys()))
        )
    url, model_name, *args = PRETRAINED_MODELS[name]
    model_path = get_data(url, model_dir)
    if model_name in ["d2v", "w2v"]:
        postfix = ".bin" if model_name == "d2v" else ".kv"
        model_path = path_append(model_path, os.path.basename(model_path) + postfix, to_str=True)
    return T2V(model_name, model_path, *args)



# coding: utf-8
# 2021/3/8 @ tongshiwei

from pprint import pformat
from typing import List, Dict
import networkx as nx
from copy import deepcopy

from .ast import str2ast, get_edges, link_variable

CONST_MATHORD = {r"\pi"}

__all__ = ["Formula", "FormulaGroup", "CONST_MATHORD", "link_formulas"]


class Formula(object):
    """
    Examples
    --------
    >>> f = Formula("x")
    >>> f
    <Formula: x>
    >>> f.ast
    [{'val': {'id': 0, 'type': 'mathord', 'text': 'x', 'role': None}, \
'structure': {'bro': [None, None], 'child': None, 'father': None, 'forest': None}}]
    >>> f.elements
    [{'id': 0, 'type': 'mathord', 'text': 'x', 'role': None}]
    >>> f.variable_standardization(inplace=True)
    <Formula: x>
    >>> f.elements
    [{'id': 0, 'type': 'mathord', 'text': 'x', 'role': None, 'var': 0}]
    """

    def __init__(self, formula: (str, List[Dict]), variable_standardization=False, const_mathord=None,
                 init=True, *args, **kwargs):
        """

        Parameters
        ----------
        formula: str or List[Dict]
            latex formula string or the parsed abstracted syntax tree
        variable_standardization
        const_mathord
        init
        args
        kwargs
        """
        self._formula = formula
        self._ast = None
        if init is True:
            self.reset_ast(
                formula_ensure_str=False,
                variable_standardization=variable_standardization,
                const_mathord=const_mathord, *args, **kwargs
            )

    def variable_standardization(self, inplace=False, const_mathord=None, variable_connect_dict=None):
        const_mathord = const_mathord if const_mathord is not None else CONST_MATHORD
        ast_tree = self._ast if inplace else deepcopy(self._ast)
        var_code = variable_connect_dict["var_code"] if variable_connect_dict is not None else {}
        for node in ast_tree:
            if node["val"]["type"] == "mathord":
                var = node["val"]["text"]
                if var in const_mathord:
                    continue
                else:
                    if var not in var_code:
                        var_code[var] = len(var_code)
                    node["val"]["var"] = var_code[var]
        if inplace:
            return self
        else:
            return Formula(ast_tree, is_str=False)

    @property
    def ast(self):
        return self._ast

    @property
    def elements(self):
        return [self.ast_graph.nodes[node] for node in self.ast_graph.nodes]

    @property
    def ast_graph(self) -> (nx.Graph, nx.DiGraph):
        edges = [(edge[0], edge[1]) for edge in get_edges(self._ast) if edge[2] == 3]
        tree = nx.DiGraph()
        for node in self._ast:
            tree.add_node(
                node["val"]["id"],
                **node["val"]
            )
        tree.add_edges_from(edges)
        return tree

    def to_str(self):
        return pformat(self._ast)

    def __repr__(self):
        if isinstance(self._formula, str):
            return "<Formula: %s>" % self._formula
        else:
            return super(Formula, self).__repr__()

    def reset_ast(self, formula_ensure_str: bool = True, variable_standardization=False, const_mathord=None, *args,
                  **kwargs):
        if formula_ensure_str is True and self.resetable is False:
            raise TypeError("formula must be str, now is %s" % type(self._formula))
        self._ast = str2ast(self._formula, *args, **kwargs) if isinstance(self._formula, str) else self._formula
        if variable_standardization:
            const_mathord = CONST_MATHORD if const_mathord is None else const_mathord
            self.variable_standardization(inplace=True, const_mathord=const_mathord)
        return self._ast

    @property
    def resetable(self):
        return isinstance(self._formula, str)


class FormulaGroup(object):
    """
    Examples
    ---------
    >>> fg = FormulaGroup(["x + y", "y + x", "z + x"])
    >>> fg
    <FormulaGroup: <Formula: x + y>;<Formula: y + x>;<Formula: z + x>>
    >>> fg = FormulaGroup(["x + y", Formula("y + x"), "z + x"])
    >>> fg
    <FormulaGroup: <Formula: x + y>;<Formula: y + x>;<Formula: z + x>>
    >>> fg = FormulaGroup(["x", Formula("y"), "x"])
    >>> fg.elements
    [{'id': 0, 'type': 'mathord', 'text': 'x', 'role': None}, {'id': 1, 'type': 'mathord', 'text': 'y', 'role': None},\
 {'id': 2, 'type': 'mathord', 'text': 'x', 'role': None}]
    >>> fg = FormulaGroup(["x", Formula("y"), "x"], variable_standardization=True)
    >>> fg.elements
    [{'id': 0, 'type': 'mathord', 'text': 'x', 'role': None, 'var': 0}, \
{'id': 1, 'type': 'mathord', 'text': 'y', 'role': None, 'var': 1}, \
{'id': 2, 'type': 'mathord', 'text': 'x', 'role': None, 'var': 0}]
    """

    def __init__(self,
                 formula_list: (list, List[str], List[Formula]),
                 variable_standardization=False,
                 const_mathord=None,
                 detach=True
                 ):
        forest = []
        self._formulas = []
        for formula in formula_list:
            if isinstance(formula, str):
                formula = Formula(
                    formula,
                    forest_begin=len(forest),
                )
                self._formulas.append(formula)
                tree = formula.ast
            elif isinstance(formula, Formula):
                if detach:
                    formula = deepcopy(formula)
                tree = formula.reset_ast(
                    formula_ensure_str=True,
                    variable_standardization=False,
                    forest_begin=len(forest),
                )
                self._formulas.append(formula)
            else:
                raise TypeError(
                    "the element in formula_list should be either str or Formula, now is %s" % type(Formula)
                )
            forest += tree
        variable_connect_dict = link_variable(forest)
        self._forest = forest
        if variable_standardization:
            self.variable_standardization(
                inplace=True,
                const_mathord=const_mathord,
                variable_connect_dict=variable_connect_dict
            )

    def __iter__(self):
        return iter(self._formulas)

    def __getitem__(self, item) -> Formula:
        return self._formulas[item]

    def __contains__(self, item) -> bool:
        return item in self._formulas

    def variable_standardization(self, inplace=False, const_mathord=None, variable_connect_dict=None):
        ret = []
        for formula in self._formulas:
            ret.append(formula.variable_standardization(inplace=inplace, const_mathord=const_mathord,
                                                        variable_connect_dict=variable_connect_dict))
        return ret

    def to_str(self):
        return pformat(self._forest)

    def __repr__(self):
        return "<FormulaGroup: %s>" % ";".join([repr(_formula) for _formula in self._formulas])

    @property
    def ast(self):
        return self._forest

    @property
    def elements(self):
        return [self.ast_graph.nodes[node] for node in self.ast_graph.nodes]

    @property
    def ast_graph(self) -> (nx.Graph, nx.DiGraph):
        edges = [(edge[0], edge[1]) for edge in get_edges(self._forest) if edge[2] == 3]
        tree = nx.DiGraph()
        for node in self._forest:
            tree.add_node(
                node["val"]["id"],
                **node["val"]
            )
        tree.add_edges_from(edges)
        return tree


def link_formulas(*formula: Formula, link_vars=True, **kwargs):
    forest = []
    for form in formula:
        forest += form.reset_ast(
            forest_begin=len(forest),
            **kwargs
        )
    variable_connect_dict = link_variable(forest) if link_vars else None
    for form in formula:
        form.variable_standardization(inplace=True, variable_connect_dict=variable_connect_dict, **kwargs)



# coding: utf-8
# 2021/5/16 @ tongshiwei

import traceback
import warnings
from .segment import seg
from .tokenization import tokenize, link_formulas
from .parser import Parser

__all__ = ["is_sif", "to_sif", "sif4sci"]


def is_sif(item):
    r"""
    Parameters
    ----------
    item

    Returns
    -------
    when item can not be parsed correctly, raise Error;
    when item doesn't need to be modified, return Ture;
    when item needs to be modified, return False;

    Examples
    --------
    >>> text = '若$x,y$满足约束条件' \
    ...        '$\\left\\{\\begin{array}{c}2 x+y-2 \\leq 0 \\\\ x-y-1 \\geq 0 \\\\ y+1 \\geq 0\\end{array}\\right.$，' \
    ...        '则$z=x+7 y$的最大值$\\SIFUnderline$'
    >>> is_sif(text)
    True
    >>> text = '某校一个课外学习小组为研究某作物的发芽率y和温度x（单位...'
    >>> is_sif(text)
    False
    """
    item_parser = Parser(item)
    item_parser.description_list()
    if item_parser.fomula_illegal_flag:
        raise ValueError(item_parser.fomula_illegal_message)
    if item_parser.error_flag == 0 and item_parser.modify_flag == 0:
        return True
    return False


def to_sif(item):
    r"""
    Parameters
    ----------
    item

    Returns
    -------
    item

    Examples
    --------
    >>> text = '某校一个课外学习小组为研究某作物的发芽率y和温度x（单位...'
    >>> siftext = to_sif(text)
    >>> siftext
    '某校一个课外学习小组为研究某作物的发芽率$y$和温度$x$（单位...'
    """
    item_parser = Parser(item)
    item_parser.description_list()
    item = item_parser.text
    return item


def sif4sci(item: str, figures: (dict, bool) = None, safe=True, symbol: str = None, tokenization=True,
            tokenization_params=None, errors="raise"):
    r"""

    Default to use linear Tokenizer, change the tokenizer by specifying tokenization_params

    Parameters
    ----------
    item
    figures
    safe
    symbol
    tokenization
    tokenization_params:
        method: which tokenizer to be used, "linear" or "ast"
        The parameters only useful for "linear":

        The parameters only useful for "ast":
            ord2token: whether to transfer the variables (mathord) and constants (textord) to special tokens.
            var_numbering: whether to use number suffix to denote different variables
    errors:
        warn
        raise
        coerce
        strict
        ignore

    Returns
    -------
    When tokenization is False, return SegmentList;
    When tokenization is True, return TokenList

    Examples
    --------
    >>> test_item = r"如图所示，则$\bigtriangleup ABC$的面积是$\SIFBlank$。$\FigureID{1}$"
    >>> tl = sif4sci(test_item)
    >>> tl
    ['如图所示', '\\bigtriangleup', 'ABC', '面积', '\\SIFBlank', \FigureID{1}]
    >>> tl.describe()
    {'t': 2, 'f': 2, 'g': 1, 'm': 1}
    >>> with tl.filter('fgm'):
    ...     tl
    ['如图所示', '面积']
    >>> with tl.filter(keep='t'):
    ...     tl
    ['如图所示', '面积']
    >>> with tl.filter():
    ...     tl
    ['如图所示', '\\bigtriangleup', 'ABC', '面积', '\\SIFBlank', \FigureID{1}]
    >>> tl.text_tokens
    ['如图所示', '面积']
    >>> tl.formula_tokens
    ['\\bigtriangleup', 'ABC']
    >>> tl.figure_tokens
    [\FigureID{1}]
    >>> tl.ques_mark_tokens
    ['\\SIFBlank']
    >>> sif4sci(test_item, symbol="gm", tokenization_params={"formula_params": {"method": "ast"}})
    ['如图所示', <Formula: \bigtriangleup ABC>, '面积', '[MARK]', '[FIGURE]']
    >>> sif4sci(test_item, symbol="tfgm")
    ['[TEXT]', '[FORMULA]', '[TEXT]', '[MARK]', '[TEXT]', '[FIGURE]']
    >>> sif4sci(test_item, symbol="gm",
    ... tokenization_params={"formula_params": {"method": "ast", "return_type": "list"}})
    ['如图所示', '\\bigtriangleup', 'A', 'B', 'C', '面积', '[MARK]', '[FIGURE]']
    >>> test_item_1 = {
    ...     "stem": r"若$x=2$, $y=\sqrt{x}$，则下列说法正确的是$\SIFChoice$",
    ...     "options": [r"$x < y$", r"$y = x$", r"$y < x$"]
    ... }
    >>> tls = [
    ...     sif4sci(e, symbol="gm",
    ...     tokenization_params={
    ...         "formula_params": {
    ...             "method": "ast", "return_type": "list", "ord2token": True, "var_numbering": True,
    ...             "link_variable": False}
    ...     })
    ...     for e in ([test_item_1["stem"]] + test_item_1["options"])
    ... ]
    >>> tls[1:]
    [['mathord_0', '<', 'mathord_1'], ['mathord_0', '=', 'mathord_1'], ['mathord_0', '<', 'mathord_1']]
    >>> link_formulas(*tls)
    >>> tls[1:]
    [['mathord_0', '<', 'mathord_1'], ['mathord_1', '=', 'mathord_0'], ['mathord_1', '<', 'mathord_0']]
    >>> from EduNLP.utils import dict2str4sif
    >>> test_item_1_str = dict2str4sif(test_item_1, tag_mode="head", add_list_no_tag=False)
    >>> test_item_1_str  # doctest: +ELLIPSIS
    '$\\SIFTag{stem}$...则下列说法正确的是$\\SIFChoice$$\\SIFTag{options}$$x < y$$\\SIFSep$$y = x$$\\SIFSep$$y < x$'
    >>> tl1 = sif4sci(test_item_1_str, symbol="gm",
    ... tokenization_params={"formula_params": {"method": "ast", "return_type": "list", "ord2token": True}})
    >>> tl1.get_segments()[0]
    ['\\SIFTag{stem}']
    >>> tl1.get_segments()[1:3]
    [['[TEXT_BEGIN]', '[TEXT_END]'], ['[FORMULA_BEGIN]', 'mathord', '=', 'textord', '[FORMULA_END]']]
    >>> tl1.get_segments(add_seg_type=False)[0:3]
    [['\\SIFTag{stem}'], ['mathord', '=', 'textord'], ['mathord', '=', 'mathord', '{ }', '\\sqrt']]
    >>> test_item_2 = {"options": [r"$x < y$", r"$y = x$", r"$y < x$"]}
    >>> test_item_2
    {'options': ['$x < y$', '$y = x$', '$y < x$']}
    >>> test_item_2_str = dict2str4sif(test_item_2, tag_mode="head", add_list_no_tag=False)
    >>> test_item_2_str
    '$\\SIFTag{options}$$x < y$$\\SIFSep$$y = x$$\\SIFSep$$y < x$'
    >>> tl2 = sif4sci(test_item_2_str, symbol="gms",
    ... tokenization_params={"formula_params": {"method": "ast", "return_type": "list"}})
    >>> tl2
    ['\\SIFTag{options}', 'x', '<', 'y', '[SEP]', 'y', '=', 'x', '[SEP]', 'y', '<', 'x']
    >>> tl2.get_segments(add_seg_type=False)
    [['\\SIFTag{options}'], ['x', '<', 'y'], ['[SEP]'], ['y', '=', 'x'], ['[SEP]'], ['y', '<', 'x']]
    >>> tl2.get_segments(add_seg_type=False, drop="s")
    [['\\SIFTag{options}'], ['x', '<', 'y'], ['y', '=', 'x'], ['y', '<', 'x']]
    >>> tl3 = sif4sci(test_item_1["stem"], symbol="gs")
    >>> tl3.text_segments
    [['说法', '正确']]
    >>> tl3.formula_segments
    [['x', '=', '2'], ['y', '=', '\\sqrt', '{', 'x', '}']]
    >>> tl3.figure_segments
    []
    >>> tl3.ques_mark_segments
    [['\\SIFChoice']]
    >>> test_item_3 = r"已知$y=x$，则以下说法中$\textf{正确,b}$的是"
    >>> tl4 = sif4sci(test_item_3)
    Warning: there is some chinese characters in formula!
    >>> tl4.text_segments
    [['已知'], ['说法', '中', '正确']]
    """
    try:
        if safe is True and is_sif(item) is not True:
            item = to_sif(item)

        ret = seg(item, figures, symbol)

        if tokenization is True:
            ret = tokenize(ret, **(tokenization_params if tokenization_params is not None else {}))

        return ret
    except Exception as e:  # pragma: no cover
        msg = traceback.format_exc()
        if errors == "warn":
            warnings.warn(msg)
        elif errors == "raise":
            raise e
